from imgui_bundle import imgui

error       = imgui.ImColor.hsv(0.9667,.88,.64).value
error_bright= imgui.ImColor.hsv(0.9667,.88,.93).value
error_dark  = imgui.ImColor.hsv(0.9667,.88,.43).value

warning     = imgui.ImVec4(1.0000, 193/255, 7/255, 1.)

ok          = imgui.ImVec4(0.0000, 0.8500, 0.0000, 1.)

gray        = imgui.ImVec4(0.5000, 0.5000, 0.5000, 1.)